#include "network.h"
#include <enet.h>
#include "Logger.h"

class ENetClient : public Network {
public:
    ENetClient() : client(nullptr), peer(nullptr) {}

    void Initialize() override {
        Logger& logger = Logger::GetInstance();

        if (enet_initialize() != 0) {
            logger.LogError("An error occurred while initializing ENet.", "ENetClient");
            exit(EXIT_FAILURE);
        }

        client = enet_host_create(nullptr, 1, 2, 0, 0);
        if (!client) {
            logger.LogError("An error occurred while creating the ENet client.", "ENetClient");
            exit(EXIT_FAILURE);
        }

        ENetAddress address;
        enet_address_set_host(&address, "localhost");
        address.port = 1234;

        peer = enet_host_connect(client, &address, 2, 0);
        if (!peer) {
            logger.LogError("No available peers for initiating connection.", "ENetClient");
            exit(EXIT_FAILURE);
        }

        logger.LogInfo("ENet client initialized and connected to server.", "ENetClient");
    }

    void Shutdown() override {
        Logger& logger = Logger::GetInstance();

        if (peer) {
            enet_peer_disconnect(peer, 0);
            logger.LogInfo("Disconnected from peer.", "ENetClient");
        }
        enet_host_destroy(client);
        enet_deinitialize();
        logger.LogInfo("ENet client shut down.", "ENetClient");
    }

    void SendData(const void* data, size_t size) override {
        Logger& logger = Logger::GetInstance();

        if (peer) {
            ENetPacket* packet = enet_packet_create(data, size, ENET_PACKET_FLAG_RELIABLE);
            enet_peer_send(peer, 0, packet);
            logger.LogDebug("Data sent to server.", "ENetClient");
        }
    }

    void ReceiveData() override {
        Logger& logger = Logger::GetInstance();

        ENetEvent event;
        while (enet_host_service(client, &event, 0) > 0) {
            if (event.type == ENET_EVENT_TYPE_RECEIVE) {
                logger.LogInfo("Received packet: " + std::string(reinterpret_cast<const char*>(event.packet->data)), "ENetClient");
                enet_packet_destroy(event.packet);
            }
        }
    }

private:
    ENetHost* client;
    ENetPeer* peer;
};
